package goaml.lang.transition;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import goaml.lang.StateChart;
import goaml.lang.entity.Entity;
import goaml.lang.parameter.Constant;
import goaml.lang.parameter.Reference;
import goaml.lang.state.AbstractState;
import goaml.lang.state.CompositeState;
import goaml.lang.state.SequentialState;
import goaml.lang.state.State;

public abstract class AbstractTransition {

	protected State source;
	protected State target;

	protected AbstractTransition() {
	}

	protected AbstractTransition(final State source, final State target) {
		this.source = source;
		this.target = target;
	}

	// From ONE "Transition" tag, create every single transition
	// For example, if the source state is a Composite State,
	// this ONE tag will generate one transition of the same class
	// for every subState of the Composite
	// with the same parameters except if there are some guards
	public static final void parse(final Node node, final StateChart stateChart) {

		if (node == null) {
			throw new RuntimeException("\"node\" is null");
		}
		if (node.getNodeName() != "Transition") {
			throw new RuntimeException("NodeName \"Transition\" was expected, got : " + node.getNodeName());
		}
		if (!node.hasChildNodes()) {
			throw new RuntimeException("A Transition node must contains others nodes (parameters, etc.)");
		}

		// Source
		String sourceEntityName = null, sourceStateName = null;
		int sourceMinCohortGuard = 0, sourceMaxCohortGuard = -1;

		// Target
		String targetEntityName = null, targetStateName = null;
		int targetMinCohortGuard = 0, targetMaxCohortGuard = -1;

		// Constants
		List<Constant> constants = new ArrayList<Constant>();

		// References
		List<Reference> references = new ArrayList<Reference>();

		// Class
		String clazz;

		// =========================
		// Initialize attributes
		// =========================
		// Required
		NamedNodeMap nodeMap = node.getAttributes();
		if (nodeMap.getNamedItem("class") != null) {
			clazz = nodeMap.getNamedItem("class").getTextContent();
		} else {
			throw new RuntimeException("\"class\" attribute missed");
		}

		// For each subNodes
		for (int i = 0; i < node.getChildNodes().getLength(); i++) {
			Node subNode = node.getChildNodes().item(i);
			NamedNodeMap subNodeMap = subNode.getAttributes();

			switch (subNode.getNodeName()) {
			case "source":
				// Required
				if (subNodeMap.getNamedItem("entity") != null) {
					sourceEntityName = subNodeMap.getNamedItem("entity").getTextContent();
				} else {
					throw new RuntimeException("\"entity\" attribute missed in \"source\" tag");
				}
				if (subNodeMap.getNamedItem("state") != null) {
					sourceStateName = subNodeMap.getNamedItem("state").getTextContent();
				} else {
					throw new RuntimeException("\"state\" attribute missed in \"source\" tag");
				}

				// Optional
				if (subNodeMap.getNamedItem("minCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent()) >= 0) {
						sourceMinCohortGuard = Integer
								.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("minCohortGuard is negative in source tag");
					}
				}
				if (subNodeMap.getNamedItem("maxCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent()) >= 0) {
						sourceMaxCohortGuard = Integer
								.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("maxCohortGuard is negative in source tag");
					}
				}

				if (sourceMaxCohortGuard != -1 && sourceMinCohortGuard > sourceMaxCohortGuard) {
					throw new RuntimeException("sourceMinCohortGuard is greater than sourceMaxCohortGuard");
				}
				break;

			case "target":
				// Required
				if (subNodeMap.getNamedItem("entity") != null) {
					targetEntityName = subNodeMap.getNamedItem("entity").getTextContent();
				} else {
					throw new RuntimeException("\"entity\" attribute missed in \"target\" tag");
				}
				if (subNodeMap.getNamedItem("state") != null) {
					targetStateName = subNodeMap.getNamedItem("state").getTextContent();
				} else {
					throw new RuntimeException("\"state\" attribute missed in \"target\" tag");
				}

				// Optional
				if (subNodeMap.getNamedItem("minCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent()) >= 0) {
						targetMinCohortGuard = Integer
								.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("minCohortGuard is negative in target tag");
					}
				}
				if (subNodeMap.getNamedItem("maxCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent()) >= 0) {
						targetMaxCohortGuard = Integer
								.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("maxCohortGuard is negative in target tag");
					}
				}

				if (targetMaxCohortGuard != -1 && targetMinCohortGuard > targetMaxCohortGuard) {
					throw new RuntimeException("targetMinCohortGuard is greater than targetMaxCohortGuard");
				}

				break;

			case "Constant": // Optional
				constants.add(Constant.parse(subNode));
				break;

			case "Reference": // Optional
				references.addAll(Reference.parse(subNode, stateChart));
				break;

			default:
				throw new RuntimeException("Invalid node tag : " + subNode.getNodeName());
			}
		}

		// =========================
		// Create transitions
		// =========================
		if (sourceEntityName.equals(targetEntityName)) {

			List<Entity> entities = stateChart.getEntitiesByName(sourceEntityName);

			for (Entity sourceEntity : entities) {
				// For each entity
				// It could be more than one because severals entities can have the same name
				// because of \"numberOfInstances\"

				List<AbstractState> sourceStates = sourceEntity.getStatesByName(sourceStateName);
				List<AbstractState> targetStates = sourceEntity.getStatesByName(targetStateName);

				if (sourceStateName.equals(targetStateName)) {
					for (AbstractState source : sourceStates) {
						// Same as for entities, it's possible to have more than one state with the
						// same name in a single entity due to SequentialState which create multiple
						// cohorts of the same state
						try {
							// Get a instance of the real transition (subClass of Transition)
							AbstractTransition transition = (AbstractTransition) Class
									.forName("goaml.lang.transition." + clazz).newInstance();

							// If numCohort is not good
							if (source.getNumCohort() < sourceMinCohortGuard
									|| (sourceMaxCohortGuard != -1 && source.getNumCohort() > sourceMaxCohortGuard)) {
								continue;
							}

							// Create transition
							transition.createAtomicTransition(stateChart, source, source, constants, references,
									sourceMinCohortGuard, sourceMaxCohortGuard, targetMinCohortGuard,
									targetMaxCohortGuard, true);
						} catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
							e.printStackTrace();
							throw new RuntimeException("Failed to create the transition !");
						}
					}
				} else {
					for (AbstractState source : sourceStates) {
						// Same as for entities, it's possible to have more than one state with the
						// same name in a single entity due to SequentialState which create multiple
						// cohorts of the same state

						// If numCohort is not good
						if (source.getNumCohort() < sourceMinCohortGuard
								|| (sourceMaxCohortGuard != -1 && source.getNumCohort() > sourceMaxCohortGuard)) {
							continue;
						}
						for (AbstractState target : targetStates) {
							// Same as source states

							// If numCohort is not good
							if (target.getNumCohort() < targetMinCohortGuard
									|| (targetMaxCohortGuard != -1 && target.getNumCohort() > targetMaxCohortGuard)) {
								continue;
							}

							try {
								// Get the real transition (subClass of Transition)
								AbstractTransition transition = (AbstractTransition) Class
										.forName("goaml.lang.transition." + clazz).newInstance();

								// Create transition
								transition.createAtomicTransition(stateChart, source, target, constants, references,
										sourceMinCohortGuard, sourceMaxCohortGuard, targetMinCohortGuard,
										targetMaxCohortGuard, false);
							} catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
								e.printStackTrace();
								throw new RuntimeException("Failed to create the transition !");
							}
						}
					}
				}
			}
		} else {
			throw new RuntimeException("Source state and target state must be the same");
		}
	}

	private final void createAtomicTransition(final StateChart stateChart, final AbstractState source,
			final AbstractState target, final List<Constant> constants, final List<Reference> references,
			final int sourceMinCohortGuard, final int sourceMaxCohortGuard, final int targetMinCohortGuard,
			final int targetMaxCohortGuard, final boolean sourceEqualTarget) {

		if (source instanceof CompositeState) {
			for (AbstractState subSource : ((CompositeState) source).getStates()) {
				if (sourceEqualTarget) {
					// source = subSource, target = subSource
					createAtomicTransition(stateChart, subSource, subSource, constants, references, sourceMinCohortGuard,
							sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, true);
				} else {
					createAtomicTransition(stateChart, subSource, target, constants, references, sourceMinCohortGuard,
							sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, false);
				}
			}
			return;
		} else if (source instanceof SequentialState) {
			AbstractState[] cohorts = ((SequentialState) source).getCohorts();
			for (int i = 0; i < cohorts.length; i++) {
				if (sourceEqualTarget) {
					// source = cohorts[i], target = cohorts[i]
					createAtomicTransition(stateChart, cohorts[i], cohorts[i], constants, references, sourceMinCohortGuard,
							sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, true);
				} else {
					createAtomicTransition(stateChart, cohorts[i], target, constants, references, sourceMinCohortGuard,
							sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, false);
				}
			}
			return;
		} else if (target instanceof CompositeState) {
			for (AbstractState subTarget : ((CompositeState) target).getStates()) {
				createAtomicTransition(stateChart, source, subTarget, constants, references, sourceMinCohortGuard,
						sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, false);
			}
			return;
		} else if (target instanceof SequentialState) {
			AbstractState[] cohorts = ((SequentialState) target).getCohorts();
			for (int i = 0; i < cohorts.length; i++) {
				createAtomicTransition(stateChart, source, cohorts[i], constants, references, sourceMinCohortGuard,
						sourceMaxCohortGuard, targetMinCohortGuard, targetMaxCohortGuard, false);
			}
			return;
		} else if (source instanceof State && target instanceof State) {
			// SubClasses know how to initialize themselves
			initialize(stateChart, (State) source, (State) target, constants, references);

		} else {
			throw new RuntimeException("Invalid source and/or target, got source : " + source.getClass() + ", target : "
					+ target.getClass());
		}
	}

	protected abstract void initialize(final StateChart stateChart, final State source, final State target,
			final List<Constant> constants, final List<Reference> references);

	// Create DelayedUpdates in order to apply computed value on the states when all transitions
	// and processes have compute
	public abstract void compute(final StateChart stateChart);

}
